# frozen_string_literal: true

module Hyrax
  module Actors
    # Actions for a file identified by file_set and relation (maps to use predicate)
    # @note Spawns asynchronous jobs
    class FileActor
      attr_reader :file_set, :relation, :user, :use_valkyrie
      include HyraxHelper
      # @param [FileSet] file_set the parent FileSet
      # @param [Symbol, #to_sym] relation the type/use for the file
      # @param [User] user the user to record as the Agent acting upon the file
      def initialize(file_set, relation, user)
        @file_set = file_set
        @relation = normalize_relation(relation)
        @user = user
      end

      # Persists file as part of file_set and spawns async job to characterize and create derivatives.
      # @param [JobIoWrapper] io the file to save in the repository, with mime_type and original_name
      # @return [CharacterizeJob, FalseClass] spawned job on success, false on failure
      # @note Instead of calling this method, use IngestJob to avoid synchronous execution cost
      # @see IngestJob
      # @todo create a job to monitor the temp directory (or in a multi-worker system, directories!) to prune old files that have made it into the repo
      def ingest_file(io)
        io.uploaded_file.update(file_set_uri: file_set.uri.to_s) if io.uploaded_file.file_set_uri.blank?

        s3_file_url = io.uploaded_file.file_url
        Hydra::Works::AddExternalFileToFileSet.call(file_set,
                                            s3_file_url,
                                            relation,
                                            {versioning: false})
        return false unless file_set.save

        repository_file = related_file

        create_version(repository_file, user)
        set_characterization_data(s3_file_url, file_set)
      end

      # Reverts file and spawns async job to characterize and create derivatives.
      # @param [String] revision_id
      # @return [CharacterizeJob, FalseClass] spawned job on success, false on failure
      def revert_to(revision_id)
        repository_file = related_file
        repository_file.restore_version(revision_id)
        return false unless file_set.save

        create_version(repository_file, user)
        CharacterizeJob.perform_later(file_set, repository_file.id)
      end

      # @note FileSet comparison is limited to IDs, but this should be sufficient, given that
      #   most operations here are on the other side of async retrieval in Jobs (based solely on ID).
      def ==(other)
        return false unless other.is_a?(self.class)

        file_set.id == other.file_set.id && relation == other.relation && user == other.user
      end

      private

      ##
      # Wraps the verisoning service with erro handling. if the service's
      # create handler isn't implemented, we want to accept that quietly here.
      def create_version(content, user)
        Hyrax::VersioningService.create(content, user)
      rescue NotImplementedError
        :no_op
      end

      ## using this method to set the characterization data instead of running the
      # actual characterization service
      def set_characterization_data(s3_file_url, file_set)
        path = URI.parse(Addressable::URI.unencode(s3_file_url)).path
        file_name = path.split('/')[-1].downcase
        uploaded_file_id = path.split('/')[-2]
        uploaded_file = Hyrax::UploadedFile.find(uploaded_file_id)
        
        parent_object = file_set.parent_object
        siblings = parent_object.file_sets

        duplicate_record = siblings.select {|fs| fs.title[0].downcase == file_name && fs.id != file_set.id }
        new_title = nil

        if duplicate_record.present?
          new_title = genrate_file_set_title(siblings, file_name)
        end

        file_set.title = [new_title] if new_title.present?
        file_set.label = new_title if new_title.present?
        file_set.characterization_proxy.file_name = new_title || file_name
        file_set.characterization_proxy.original_name = new_title || file_name
        file_set.characterization_proxy.mime_type = uploaded_file.characterization_data['content_type']
        file_set.characterization_proxy.file_size = uploaded_file.characterization_data['file_size']
        file_set.characterization_proxy.original_checksum = uploaded_file.characterization_data['original_checksum']
        file_set.characterization_proxy.format_label = []

        if file_set.mime_type_image?
          file_set.characterization_proxy.height = uploaded_file.characterization_data['height']
          file_set.characterization_proxy.width  = uploaded_file.characterization_data['width']

          file_set.characterization_proxy.alpha_channels = channels(s3_file_url) if file_set.image?
        end

        new_url = move_s3_object_to_work_bucket(uploaded_file, file_set)

        file_set.characterization_proxy.mime_type = "message/external-body; access-type=URL; URL=\"#{new_url}\""

        file_set.characterization_proxy.save!

        file_set.label = file_set.characterization_proxy.original_name

        file_set.save!
      end

      ##
      # @return [Hydra::PCDM::File] the file referenced by relation
      def related_file
        file_set.public_send(normalize_relation(relation)) || raise("No #{relation} returned for FileSet #{file_set.id}")
      end

      def normalize_relation(relation)
        use_valkyrie ? normalize_relation_for_valkyrie(relation) : normalize_relation_for_active_fedora(relation)
      end

      def normalize_relation_for_active_fedora(relation)
        return relation.to_sym if relation.respond_to? :to_sym

        case relation
        when Hyrax::FileMetadata::Use::ORIGINAL_FILE
          :original_file
        when Hyrax::FileMetadata::Use::EXTRACTED_TEXT
          :extracted_file
        when Hyrax::FileMetadata::Use::THUMBNAIL
          :thumbnail_file
        else
          :original_file
        end
      end

      ##
      # @return [RDF::URI]
      def normalize_relation_for_valkyrie(relation)
        return relation if relation.is_a?(RDF::URI)

        Hyrax::FileMetadata::Use.uri_for(use: relation.to_sym)
      rescue ArgumentError
        Hyrax::FileMetadata::Use::ORIGINAL_FILE
      end

      def pathhint(io)
        io.uploaded_file&.uploader&.path || io.path
      end

      def move_s3_object_to_work_bucket(uploaded_file, file_set)
        s3 = S3StorageService.new
        s3.init_client

        source_object_key = uploaded_file.file.path
        target_bucket_name = s3.sanitise_name(file_set.parent_works.first.id)

        target_bucket = Aws::S3::Bucket.new(target_bucket_name)

        s3.create_bucket(target_bucket_name) unless s3.bucket_exists?(target_bucket_name)

        s3.move_object(source_object_key, target_bucket_name, file_set.object_key_on_s3, { uploaded_file_path: source_object_key })

        s3.presigned_url_for_download(target_bucket_name, file_set.object_key_on_s3)
      end

      def clear_metadata(file_set)
        # The characterization of additional file versions adds new height/width/size/checksum values to un-orderable...
        # `ActiveTriples::Relation` fields on `original_file`. Values from those are then randomly pulled into Solr...
        # fields which may have scalar or vector cardinality. So for height/width you get two scalar values pulled from...
        # "randomized parallel arrays". Upshot is to reset all of these before (re)characterization to stop the mayhem.
        file_set.characterization_proxy.height = []
        file_set.characterization_proxy.width  = []
        file_set.characterization_proxy.original_checksum = []
        file_set.characterization_proxy.file_size = []
        file_set.characterization_proxy.format_label = []
      end

      def channels(filepath)
        ch = MiniMagick::Tool::Identify.new do |cmd|
          cmd.format '%[channels]'
          cmd << filepath
        end
        [ch]
      end

      def genrate_file_set_title(siblings, file_name)
        existing_titles = siblings.map{ |fs| fs.title[0].downcase }

        base_name, extension = File.basename(file_name, ".*"), File.extname(file_name)

        suffix = 1
        new_title = file_name
        while existing_titles.include?(new_title)
          new_title = "#{base_name}_#{suffix}#{extension}"
          suffix += 1
        end

        new_title
      end
    end
  end
end
